from tree import Tree
import random
import fitness
ALL_PRODUCTIONS_STRING=\
'''
# P means picture 
P ::= E255

# E255 means an expression whose value will be in the range 0 to 255.
E255 ::= 255 * E1P + E1P
#E255 ::= EI mod 256
E255 ::= if COND then E255 else E255
E255 ::= hsv( E1P , E1P , E1P )
E255 ::= rgb( E1P * 255 , E1P * 255 , E1P * 255 )

# E1P means an expression is a non-negative number, with maximum value 1.
E1P ::= x / w
E1P ::= y / h
E1P ::= ( sin(x) ) * ( cos(y) )
E1P ::= ( cos(x) ) * ( sin(y) )
E1P ::= ( tan(x) ) * ( tan(y) )
E1P ::= log(x)
E1P ::= log(y)
E1P ::= abs ( EE )
E1P ::= ( 1 + EE ) / 2

# COND stands for a conditional (True or False) expression.
COND ::= EI > EI
COND ::= E1P < E1P
COND ::= EE < EE
COND ::= EI > EE
COND ::= E1P < EE
COND ::= E1P = EE
COND ::= COND and COND
COND ::= not COND
COND ::= COND or COND

# EE stands for an expression whose value is in the range -1 to 1.
EE ::= ( EI mod 3 ) - 1
EE ::= sin(x)
EE ::= cos(x)
EE ::= sin(y)
EE ::= cos(y)

# EI stands for an expression whose value is any integer.
#EI ::= ceil(x)
#EI ::= floor(x)
#EI ::= ceil(y)
#EI ::= floor(y)
EI ::= x
EI ::= y
EI ::= 255
EI ::= 63
EI ::= 1
EI ::= tan(x)
EI ::= tan(y)
EI ::= log(x)
EI ::= log(y)
EI ::= EI + EI
EI ::= EI bxor EI
'''

def LHS(production):
  '''Return the nonterminal on the Left-Hand Side of the production.'''
  return production.split(" ::= ")[0]

def RHS(production):
  '''Return the phrase on the Right-Hand Side of the production.'''
  return production.split(" ::= ")[1]

PRODUCTION_STRINGS = ALL_PRODUCTIONS_STRING.split("\n")
PRODUCTION_STRINGS = [p for p in PRODUCTION_STRINGS if not ( p=="" or p[0]=='#')]
PRODUCTIONS = [(LHS(p), RHS(p).split(" ")) for p in PRODUCTION_STRINGS]
# Now each production is a tuple (pair), where the 2nd element is
# a list of the RHS items.

def apply_production(production, sentential_form, occurrence_number):
  '''Take a production (a pair), a sentential form (a list of strings), 
  and an integer i,  and return a new sentential form whose ith occurrence
  of the left-hand side of the production has been replaced by
  the right-hand side of that production.'''
  nonterminal = production[0]
  output_items = []
  special_item_count = 0
  try:
    for item in sentential_form:
      if nonterminal == item:
         if special_item_count == occurrence_number:
           for new_item in production[1]:
             output_items.append(new_item)
         else:
           output_items.append(item)
           special_item_count += 1
      else:
         output_items.append(item)
  except:
    raise Exception("There are not "+str(occurence_number + 1)+" occurrence of "+nonterminal+\
      " in "+sentential_form+".")
#  return " ".join(output_items)
  return output_items
      

def count_occurrences(production, sentential_form):
  '''Return the number of times the nonterminal symbol on the
  left-hand side of the production occurs in the sentential form.'''
  nonterminal = production[0]
  return sentential_form.count(nonterminal)

# Dictionary giving all RHS's for productions whose LHS is a certin
# NONTERMINAL:
RHSS_FOR_NONTERMINAL = {}

for prod in PRODUCTIONS:
  #print("PROD: ", prod)
  LHS = prod[0]
  RHS = prod[1]
  #print("LHS ", LHS, " RHS: ", RHS)
  try:
    old = RHSS_FOR_NONTERMINAL[LHS]

  except:
    old = []

  RHSS_FOR_NONTERMINAL[LHS] = old + [RHS]


def is_nonterminal(item):
  return item[0].isupper() # True if item starts with a capital letter.

import random

def gen_phrase(non_terminal, node, tree, dictionary, depth=0):
  '''Return a phrase (string of nonterminal and/or terminal symbols) obtained by
  expanding the given non_terminal using the grammar.
  Choose applicable productions randomly.
  (We could choose applicable productions in a first-match ordering,
   but that will be a different experiment.)'''

  output_items = []
  non_terminal_list=non_terminal.split(" ")
  non_terminal=non_terminal_list[0]
 
  rhss = RHSS_FOR_NONTERMINAL[non_terminal]

  rhs = random.choice(rhss)

  # Now for the recursive part:
  if depth==20:
  	return (-1)
  for item in rhs:
    item2=item
    if item in dictionary.keys():
    		#adding the counts of each nonterminal to the dictionary
    		item2=item+" " + str(dictionary[item])
    		dictionary[item]=dictionary[item]+1
    #adding node to the tree	
    node2 = tree.add_node(str(item2), str(node.identifier))
   
      
    if is_nonterminal(item):
       new_items = -1
       while new_items == -1:
         new_items = gen_phrase(item2, node2, tree, dictionary, depth+1)
       output_items += new_items
    else:
       output_items += [item]

  return output_items
  

def generate_formula():
  '''Test the formula generation.'''
  #defining the dictionary of non_terminals
  non_terminal_dictionary = {"P":1, "E1P":0, "EI":0, "EE":0, "E255":0, "COND":0}
  
  #initializing tree and adding root node
  tree=Tree()
  node=tree.add_node("P")

  #calling gen_phrase, returning tree_list with tree, formula, and dictionary
  formula=gen_phrase("P", node,tree, non_terminal_dictionary)
  if formula==(-1):
  	formula=gen_phrase("P", node,tree, non_terminal_dictionary)
  fitness_value=0
  tree_list = [tree, get_leaf_nodes(tree), non_terminal_dictionary, fitness_value]
  return tree_list


def apply_mutation(tree, non_terminal_dictionary):
	'''Asks for a tree and the dictionary of how many times each 
	non terminal appears. Returns a new mutated tree, formula, and the
	updated dictionary'''

	mutated_tree=copy_tree(tree)
	random_choice_list=[]
	for i in non_terminal_dictionary.keys():
		for x in range(non_terminal_dictionary[i]):
			random_choice_list.append(i)
	random_choice_list.remove("E255")
	random_choice_list.remove("P")
	random_choice=random.choice(random_choice_list)

	extra_number=random.randint(0, non_terminal_dictionary[random_choice]-1)


	random_choice_number=random_choice+" " + str(extra_number)

	for node in mutated_tree.traverse("P"):
		for child in mutated_tree[random_choice_number].children:
			mutated_tree.remove_node(child, random_choice_number)
			

# Till here, the first tree was created, and the children of the chosen node were removed

#A new tree is defined where gen_phrase is called with the chosen 
#nonterminal. This is the replacement.
	tree2=Tree()
	node2=tree2.add_node(random_choice_number)
	
	non_terminal_dictionary2 = {"P":0, "E1P":0, "EI":0, "EE":0, "E255":0, "COND":0}
	
	formula2=gen_phrase(random_choice_number, node2, tree2, non_terminal_dictionary2)
	tree2_list=[tree2, formula2, non_terminal_dictionary2, 0]

	#adding each node from tree2 to tree
	for node in tree2.traverse(random_choice_number):
		for child in tree2[node].children:
			mutated_tree.add_node(child, node)
	

	#renaming each node because the replacement caused the naming to be off
	non_terminal_dictionary = {"P":0, "E1P":0, "EI":0, "EE":0, "E255":0, "COND":0}
	non_terminal_dictionary_copy = {"P":0, "E1P":0, "EI":0, "EE":0, "E255":0, "COND":0}
	for node in mutated_tree.traverse("P"):
		node_list=node.split(" ")
		node=node_list[0]
		if node in non_terminal_dictionary:
			non_terminal_dictionary[node]=non_terminal_dictionary[node]+1
			non_terminal_dictionary_copy[node]=non_terminal_dictionary_copy[node]+1


	for node in mutated_tree.traverse("P"):
		if (is_nonterminal(node)) and (node!="P"):
			node_list=node.split(" ")

			node_list[1]=non_terminal_dictionary[node_list[0]]-non_terminal_dictionary_copy[node_list[0]]

			non_terminal_dictionary_copy[node_list[0]]=non_terminal_dictionary_copy[node_list[0]]-1
			node=node_list[0]+" " +str(node_list[1])
			
		
	return([mutated_tree, get_leaf_nodes(mutated_tree), non_terminal_dictionary, 0])
			

def cross_over(tree1_list, tree2_list):
        '''The input is two treelists and output is a child'''
	combined_dictionary=tree1_list[2]
	keys_for_tree_1=tree1_list[2].keys()

	for i in keys_for_tree_1:
		minimum=min(tree1_list[2][i], tree2_list[2][i])
		combined_dictionary[i]=minimum
	random_choice_list=[]
	for i in combined_dictionary.keys():
		for x in range(combined_dictionary[i]):
			random_choice_list.append(i)
	random_choice_list.remove("E255")
	random_choice_list.remove("P")	
	random_choice=random.choice(random_choice_list)
	extra_number=random.randint(0, combined_dictionary[random_choice]-1)
	random_choice_number=random_choice+" " + str(extra_number)

	#up till now, choosing which non-terminal to change
	
	tree1=copy_tree(tree1_list[0])
	tree2=tree2_list[0]
	for node in tree1.traverse("P"):
		for child in tree1[random_choice_number].children:
			tree1.remove_node(child, random_choice_number)
	for node in tree2.traverse(random_choice_number):
		for child in tree2[node].children:
			tree1.add_node(child, node)
	#replacing nodes with tree2 nodes instead of random generation as in mutation
	return (update_image(tree1))

def get_leaf_nodes(tree):
	'''Returns the leaf nodes of a given tree. Used for getting formula
	from tree.'''
	nodes=""
	for node in tree.traverse("P"):
    		if tree[node].children==[]:
       		 nodes=nodes+" " +str(node)
	return nodes
			

def adding_to_list(list_with_formulas, length):
	'''Asks for an input list and how many formulas need to be added.
	Outputs a list with that many new formulas appended to it 
	using generate_formulas().'''
	
	for i in range(length):
		temp=generate_formula()
		list_with_formulas.append(temp)
	return (list_with_formulas)


def apply_fitness_function(image_list):
	'''Asks for an input list of formulas. Applies a fitness function
	using statistics programmed in the fitness file. Outputs a list with
	fitness values added'''
	for formula in image_list:				
		fitness_value=0
		RGB_values=[]
		HSV_values=[]
		window_number=pmNewComputedImage(formula[1], 256, 256, formula[1])

		#analyzing the RGB values for each pixel
		
		for y in range(255):
			for x in range(255):
				RGB_value=pmGetPixel(window_number, x, y)
				HSV_value=pmRGBtoHSV(RGB_value)
				mean_of_values=(RGB_value[0]+RGB_value[1]+RGB_value[2])/3
				RGB_values.append(mean_of_values)
				HSV_values.append(HSV_value)
		#sorting by ascending order and calling fitness functions
		RGB_values.sort(reverse=True)
		list_of_statistics=(fitness.all_stats(RGB_values))
		colors_present=fitness.colors(HSV_values)
		fitness_value=fitness_value+fitness.mean_fitness(list_of_statistics)
		fitness_value=fitness_value+fitness.median_fitness(list_of_statistics)
		fitness_value=fitness_value+fitness.most_fitness(list_of_statistics)
		fitness_value=fitness_value+fitness.standard_deviation_fitness(list_of_statistics)
		fitness_value=fitness_value+fitness.color_fitness(colors_present[1])
		formula[3]=(fitness_value)
		pmClose(window_number)
	#keeping and returning the 3 best formulas along with their info			
	return (image_list)
			
	
	
def print_list(input_list):
	'''Asks for an input list and outputs each element of the list
	in a new line'''
	for i in range (len(input_list)):
		print (input_list[i])

def copy_tree(input_tree):
	'''Copies the input tree. Used in mutation and crossover so that
	the original tree isn't harmed'''
	output_tree=Tree()
	output_tree.add_node("P")
	for node in input_tree.traverse("P"):
		for child in input_tree[node].children:
			output_tree.add_node(child, node)
	return (output_tree)

def update_image(tree):
	'''Updates the dictionary with the current counts and renames the 
	tree as to make sure it is correct'''
	non_terminal_dictionary = {"P":0, "E1P":0, "EI":0, "EE":0, "E255":0, "COND":0}
	non_terminal_dictionary_copy = {"P":0, "E1P":0, "EI":0, "EE":0, "E255":0, "COND":0}
	for node in tree.traverse("P"):
		node_list=node.split(" ")
		node=node_list[0]
		if node in non_terminal_dictionary:
			non_terminal_dictionary[node]=non_terminal_dictionary[node]+1
			non_terminal_dictionary_copy[node]=non_terminal_dictionary_copy[node]+1

	
	for node in tree.traverse("P"):
		if (is_nonterminal(node)) and (node!="P"):
			node_list=node.split(" ")
			node_list[1]=non_terminal_dictionary[node_list[0]]-non_terminal_dictionary_copy[node_list[0]]
			non_terminal_dictionary_copy[node_list[0]]=non_terminal_dictionary_copy[node_list[0]]-1
			node=node_list[0]+" " +str(node_list[1])
	return ([tree, get_leaf_nodes(tree), non_terminal_dictionary, 0])

def create_initial_population(number_of_images=20):
	'''Asks for the initial number of images and initializes the 
	population'''
	initial_population = adding_to_list([], number_of_images)
	initial_population=apply_fitness_function(initial_population)
	initial_population.sort(key=lambda x:x[3], reverse=True)
	initial_population=initial_population[0:6]
	return initial_population
	
def run_mutation(image_list):
	'''Then, a mutation is applied and a fitness function is applied. The best images from the
	non-mutated formulas and best images from the mutated formulas are kept.'''
	formula_list=[]
	mutated_list = []
	for element in image_list:
		element_list=[copy_tree(element[0]), element[1], element[2], element[3]]
		formula_list.append(element_list)
	#makes a list of original images
	for formula in formula_list:
		new_formula=apply_mutation(formula[0], formula[2])
		mutated_list.append(new_formula)
	#makes list of mutated images
	formula_list=apply_fitness_function(formula_list)
	mutated_list=apply_fitness_function(mutated_list)
	#applies fitness on both lists
	image_list_output=[]
	for i in range(len(image_list)):
		if formula_list[i][3]>mutated_list[i][3]:
			image_list_output.append(formula_list[i])
		if formula_list[i][3]<mutated_list[i][3]:
			image_list_output.append(mutated_list[i])
		if formula_list[i][3]==mutated_list[i][3]:
			random_number=random.randint(0,1)
			if random_number==1:
				image_list_output.append(mutated_list[i])
			if random_number==0:
				image_list_output.append(formula_list[i])
	#compares the mutated version and the original of each image and 
	#chooses the one with higher fitness
	return (image_list_output)
		
def run_cross_over(image_list):
	'''Then, a crossover is applied and a fitness function is applied. The best child from
        each parent.'''
	output_list=[]
	for image in image_list:
		children_list=[] #resets children_list each time
		count=0
		index=image_list.index(image)
		while count<6:
			if index!=count:
				child=cross_over(image, image_list[count])
				#pmNewComputedImage("Parent1", 256, 256, image[1])
				#print("Parent1", image[1])
				#pmNewComputedImage("Parent2", 256, 256, image_list[count][1])
				#print("Parent2", image_list[count][1])
				#pmNewComputedImage("Child", 256, 256, child[1])
				#print("child", child[1])
				children_list.append(child)
				#temporary=raw_input("Move on?")
				#runs crossover with every possible combination of parents
			count=count+1
		#applies fitness function for 6 children of each parent, chooses 1
		children_list=apply_fitness_function(children_list)
		children_list.sort(key=lambda x:x[3], reverse=True)
		output_list.append(children_list[0])
	#returns 6 images
	return (output_list)
				
				
		
def gen():
        '''Main function where everything is called'''
    	#decide the number of generations
	number_of_generations=int(raw_input("How many generations would you like to run: "))
	initial_population=create_initial_population()
        #creating initial population
	initial_population_save=[]
	for element in initial_population:
		initial_population_save.append([copy_tree(element[0]), element[1], element[2], element[3]])
        #creating save so that we can display later
	pos = 0
	for formula in initial_population:
		#position window and compute image
		wn = pmNewComputedImage(formula[1], 256, 256, formula[1])
		pmPositionWindow(wn,pos,0, 0, 0)
	
		#fix the position
		pos=pos+1000/5

	for gen_count in range (number_of_generations):
		#confirming that the user wants to continue
		asking_string=" Starting generation "+str(gen_count+1) + ". Enter 'n' to quit."
		#user_continue=raw_input(asking_string)
		user_continue = 'y'
		if gen_count!=(number_of_generations-1):
			pmClose(-1)
		if user_continue=='n':
			break
		else:
			pmClose(-1)

		#run mutations on the initial population
		mutated_image_list=run_mutation(initial_population)
	
		
		#run crossovers on the mutated image list
		cross_over_image_list = run_cross_over(mutated_image_list)

		#setting initial population to be equal to cross-over for next generation
		initial_population=[]
		for element in cross_over_image_list:
			initial_population.append([copy_tree(element[0]), element[1], element[2], element[3]])

	
		pmClose(-1)
	
		#positioning the windows and opening them up for the last generation
	xpos = 0
	ypos = 0
	width = 0
	height = 0
	for image in initial_population_save:
		original_naming_string="O:"+ str(image[1])
		wn = pmNewComputedImage(original_naming_string, 256, 256, image[1])
		pmPositionWindow(wn,xpos,ypos, width, height)
		xpos=xpos+1000/5
				
	xpos = 0
	ypos = 256

	for image in mutated_image_list:
		original_naming_string="O:"+ str(image[1])
		wn = pmNewComputedImage(original_naming_string, 256, 256, image[1])
		pmPositionWindow(wn,xpos,ypos, width, height)
		xpos=xpos+1000/5
			
				
			
	xpos = 0
	ypos = 510
			
	for image in cross_over_image_list:
		modified_naming_string="M:"+ str(image[1])
		wn=pmNewComputedImage(modified_naming_string, 256, 256, image[1])
		pmPositionWindow(wn,xpos,ypos, width, height)
		xpos=xpos+1000/5

gen()
			


